//Source & infos: https://homofaciens.de/technics-miscellaneous-electric-generator_en.htm
//Display rpm and generator voltage on a 2x16 character LCD with a Simatic IOT2020
//Compile with: gcc four-stroke-generator-SIMATIC.c -lmraa -o four-stroke-generator-SIMATIC


#include "mraa.h"
#include <string.h>
//#include "Lcd.h"




#define LCD_RS    8
#define LCD_EN    9
#define LCD_D0   10
#define LCD_D1   11
#define LCD_D2   12
#define LCD_D3   13

#define ADC0      0
#define HALL_SENSOR 15

// commands
#define LCD_CLEARDISPLAY 0x01
#define LCD_RETURNHOME 0x02
#define LCD_ENTRYMODESET 0x04
#define LCD_DISPLAYCONTROL 0x08
#define LCD_CURSORSHIFT 0x10
#define LCD_FUNCTIONSET 0x20
#define LCD_SETCGRAMADDR 0x40
#define LCD_SETDDRAMADDR 0x80

// flags for display entry mode
#define LCD_ENTRYRIGHT 0x00
#define LCD_ENTRYLEFT 0x02
#define LCD_ENTRYSHIFTINCREMENT 0x01
#define LCD_ENTRYSHIFTDECREMENT 0x00

// flags for display on/off control
#define LCD_DISPLAYON 0x04
#define LCD_DISPLAYOFF 0x00
#define LCD_CURSORON 0x02
#define LCD_CURSOROFF 0x00
#define LCD_BLINKON 0x01
#define LCD_BLINKOFF 0x00

// flags for display/cursor shift
#define LCD_DISPLAYMOVE 0x08
#define LCD_CURSORMOVE 0x00
#define LCD_MOVERIGHT 0x04
#define LCD_MOVELEFT 0x00

// flags for function set
#define LCD_8BITMODE 0x10
#define LCD_4BITMODE 0x00
#define LCD_2LINE 0x08
#define LCD_1LINE 0x00
#define LCD_5x10DOTS 0x04
#define LCD_5x8DOTS 0x00


struct timespec btime, etime;__time_t i;

mraa_result_t r = MRAA_SUCCESS;

mraa_gpio_context lcd_rs;
mraa_gpio_context lcd_en;
mraa_gpio_context lcd_d0;
mraa_gpio_context lcd_d1;
mraa_gpio_context lcd_d2;
mraa_gpio_context lcd_d3;
mraa_gpio_context lcd_d4;
mraa_gpio_context lcd_d5;
mraa_gpio_context lcd_d6;
mraa_gpio_context lcd_d7;

mraa_aio_context adc0;
mraa_gpio_context hall_sensor;

uint32_t lcd_bitmode = LCD_4BITMODE;
uint32_t numlines = LCD_2LINE;
uint32_t currline;
uint32_t displayfunction;
uint32_t displaycontrol;
uint32_t displaymode;

time_t programStartSec;

void Lcd_pulseEnable(void) {
  r = mraa_gpio_write(lcd_en, 0);
  usleep(1);
  r = mraa_gpio_write(lcd_en, 1);
  usleep(1);    // enable pulse must be >450ns
  r = mraa_gpio_write(lcd_en, 0);
  usleep(100);   // commands need > 37us to settle
}


void Lcd_write4bits(uint32_t value) {
  uint32_t i;

  r = mraa_gpio_write(lcd_d0, value >> 0 & 0x01);
  r = mraa_gpio_write(lcd_d1, value >> 1 & 0x01);
  r = mraa_gpio_write(lcd_d2, value >> 2 & 0x01);
  r = mraa_gpio_write(lcd_d3, value >> 3 & 0x01);

  Lcd_pulseEnable();
}

void Lcd_write8bits(uint32_t value) {
  int i;

  r = mraa_gpio_write(lcd_d0, value >> 0 & 0x01);
  r = mraa_gpio_write(lcd_d1, value >> 1 & 0x01);
  r = mraa_gpio_write(lcd_d2, value >> 2 & 0x01);
  r = mraa_gpio_write(lcd_d3, value >> 3 & 0x01);
  r = mraa_gpio_write(lcd_d4, value >> 0 & 0x01);
  r = mraa_gpio_write(lcd_d5, value >> 1 & 0x01);
  r = mraa_gpio_write(lcd_d6, value >> 2 & 0x01);
  r = mraa_gpio_write(lcd_d7, value >> 3 & 0x01);

  Lcd_pulseEnable();
}


void Lcd_send(uint32_t value, uint32_t mode) {
  r = mraa_gpio_write(lcd_rs, mode);

  // rw pin removed, add later if needed

  if (displayfunction & LCD_8BITMODE) {
    Lcd_write8bits(value);
  } else {
    Lcd_write4bits(value>>4);
    Lcd_write4bits(value);
  }
}


void Lcd_command(uint32_t value) {
  Lcd_send(value, 0);
}

void Lcd_display() {
  displaycontrol |= LCD_DISPLAYON;
  Lcd_command(LCD_DISPLAYCONTROL | displaycontrol);
}

void Lcd_clear() {
  Lcd_command(LCD_CLEARDISPLAY);  // clear display, set cursor position to zero
  usleep(2000);  // this command takes a long time!
}


void Lcd_begin(uint32_t cols, uint32_t lines, uint32_t dotsize) {
  if (lines > 1) {
    displayfunction |= LCD_2LINE;
  }
  numlines = lines;
  currline = 0;

  // for some 1 line displays you can select a 10 pixel high font
  if ((dotsize != 0) && (lines == 1)) {
    displayfunction |= LCD_5x10DOTS;
  }

  // SEE PAGE 45/46 FOR INITIALIZATION SPECIFICATION!
  // according to datasheet, we need at least 40ms after power rises above 2.7V
  // before sending commands. Arduino can turn on way befer 4.5V so we'll wait 50
  usleep(50000);

  // Now we pull both RS and R/W low to begin commands
  r = mraa_gpio_write(lcd_rs, 0);
  r = mraa_gpio_write(lcd_en, 0);
  // no rw pin unless needed

  //put the LCD into 4 bit or 8 bit mode
  if (! (displayfunction & LCD_8BITMODE)) {
    // this is according to the hitachi HD44780 datasheet
    // figure 24, pg 46

    // we start in 8bit mode, try to set 4 bit mode
    Lcd_write4bits(0x03);
    usleep(4500); // wait min 4.1ms

    // second try
    Lcd_write4bits(0x03);
    usleep(4500); // wait min 4.1ms

    // third go!
    Lcd_write4bits(0x03);
    usleep(150);

    // finally, set to 4-bit interface
    Lcd_write4bits(0x02);
  } else {
    // this is according to the hitachi HD44780 datasheet
    // page 45 figure 23

    // Send function set command sequence
    Lcd_command(LCD_FUNCTIONSET | displayfunction);
    usleep(4500);  // wait more than 4.1ms

    // second try
    Lcd_command(LCD_FUNCTIONSET | displayfunction);
    usleep(150);

    // third go
    Lcd_command(LCD_FUNCTIONSET | displayfunction);
  }

  // finally, set # lines, font size, etc.
  Lcd_command(LCD_FUNCTIONSET | displayfunction);

  // turn the display on with no cursor or blinking default
  displaycontrol = LCD_DISPLAYON | LCD_CURSOROFF | LCD_BLINKOFF;
  Lcd_display();

  // clear it off
  Lcd_clear();

  // Initialize to default text direction (for romance languages)
  displaymode = LCD_ENTRYLEFT | LCD_ENTRYSHIFTDECREMENT;
  // set the entry mode
  Lcd_command(LCD_ENTRYMODESET | displaymode);



}


void Lcd_setCursor(uint32_t col, uint32_t row) {

  uint32_t row_offsets[] = { 0x00, 0x40, 0x14, 0x54 };
  if ( row >= numlines ) {
    row = numlines-1;    // we count rows starting w/0
  }

  Lcd_command(LCD_SETDDRAMADDR | (col + row_offsets[row]));

}

uint32_t Lcd_lcdWrite(uint32_t value) {
  Lcd_send(value, 1);
  return 1; // assume sucess
}


void Lcd_print(char input[100]) {
  int i;
  int ende = strlen(input);

  if(ende > 16){
    ende = 16;
  }
  for(i = 0; i < ende; ++i) {
    Lcd_lcdWrite(input[i]);
  }
}

void sensorTriggered(double bounceLoop) {
  int triggerPause = 0;

  triggerPause = 0;
  while(triggerPause < bounceLoop){
    if(mraa_gpio_read(hall_sensor) == 1){
      triggerPause++;
      //printf("HIGH\r");
    }
    else{
      triggerPause = 0;
      //printf("LOW\r");
    }
    usleep(10);
  }

  triggerPause = 0;
  while(triggerPause < bounceLoop){
    if(mraa_gpio_read(hall_sensor) == 0){
      triggerPause++;
      //printf("LOW\r");
    }
    else{
      triggerPause = 0;
      //printf("HIGH\r");
    }
    usleep(10);
  }

}

int get_rev_speed(long wait_time){
  int rpm = 0;
  int pulses = 0;
  time_t timeStart, timeNow = 0;

  timeStart = time(NULL);

  while(timeNow-timeStart < wait_time){
    sensorTriggered(10);
    pulses++;
    timeNow = time(NULL);
    //printf("timeNow-timeStart=%ld\n", timeNow-timeStart);
  }

  rpm = pulses * 60 / wait_time;

  //printf("pulses=%d, rpm=%d\n", pulses, rpm);

  return rpm;
}

int main(void) {
  //cout << "Hello IOT2000." << endl;
  int i;
  int rpm=0;
  char LCD_line_1[20];
  char LCD_line_2[20];
  float ADC0_value;
  double voltage;

  struct timespec ttime;

  clock_gettime(CLOCK_REALTIME, &ttime);
  programStartSec = (double)(ttime.tv_sec);

  mraa_init();

  adc0 = mraa_aio_init(ADC0);

  hall_sensor = mraa_gpio_init(HALL_SENSOR);

  lcd_rs = mraa_gpio_init(LCD_RS);
  lcd_en = mraa_gpio_init(LCD_EN);
  lcd_d0 = mraa_gpio_init(LCD_D0);
  lcd_d1 = mraa_gpio_init(LCD_D1);
  lcd_d2 = mraa_gpio_init(LCD_D2);
  lcd_d3 = mraa_gpio_init(LCD_D3);

  r = mraa_gpio_dir(hall_sensor, MRAA_GPIO_IN);
  if (r != MRAA_SUCCESS) {
    mraa_result_print(r);
  }
  r = mraa_gpio_dir(hall_sensor, MRAA_GPIO_PULLUP);
  if (r != MRAA_SUCCESS) {
    mraa_result_print(r);
  }

/*
  r = mraa_aio_dir(adc0, MRAA_AIO_IN);
  if (r != MRAA_SUCCESS) {
    mraa_result_print(r);
  }
*/

  r = mraa_gpio_dir(lcd_rs, MRAA_GPIO_OUT);
  if (r != MRAA_SUCCESS) {
    mraa_result_print(r);
  }
  r = mraa_gpio_dir(lcd_en, MRAA_GPIO_OUT);
  if (r != MRAA_SUCCESS) {
    mraa_result_print(r);
  }
  r = mraa_gpio_dir(lcd_d0, MRAA_GPIO_OUT);
  if (r != MRAA_SUCCESS) {
    mraa_result_print(r);
  }
  r = mraa_gpio_dir(lcd_d1, MRAA_GPIO_OUT);
  if (r != MRAA_SUCCESS) {
    mraa_result_print(r);
  }
  r = mraa_gpio_dir(lcd_d2, MRAA_GPIO_OUT);
  if (r != MRAA_SUCCESS) {
    mraa_result_print(r);
  }
  r = mraa_gpio_dir(lcd_d3, MRAA_GPIO_OUT);
  if (r != MRAA_SUCCESS) {
    mraa_result_print(r);
  }


  Lcd_begin(16,2, 0);
  Lcd_clear();

  Lcd_setCursor(0,0); // move to the beginning of the first line


  Lcd_print("SIMATIC IOT2020 ");
  sleep(2);


  while(1){
    ADC0_value = mraa_aio_read(adc0);
    voltage=(27.0+5.6)/5.6 * 5.0*(double)(ADC0_value)/1023.0;
    sprintf(LCD_line_1, "Crank =% 6drpm",  rpm);
    Lcd_setCursor(0,0);
    Lcd_print(LCD_line_1);
    sprintf(LCD_line_2, "Output=%6.1lfV  ",  voltage);
    //sprintf(LCD_line_2, "Sensor1=%d      ",  mraa_gpio_read(hall_sensor));
    //printf("%s ADC0_value=%lf\n",LCD_line_2, ADC0_value);
    Lcd_setCursor(0,1);
    Lcd_print(LCD_line_2);
    rpm = get_rev_speed(3);
    //sleep(1);
  }

  return MRAA_SUCCESS;
}

